@@ -20,7 +20,7 @@ let pp_error fmt e =
 
 let rev_string buf =
   let len = String.length buf in
-  let res = Bytes.make len '\000' in
+  let res = Bytes.create len in
   for i = 0 to len - 1 do
     Bytes.set res (len - 1 - i) (String.get buf i)
   done ;
@@ -28,39 +28,37 @@ let rev_string buf =
 
 exception Message_too_long
 
-let string_get_uint8 buf idx =
-  (* TODO: use String.get_uint8 when mirage-crypto-ec requires OCaml >= 4.13 *)
-  Bytes.get_uint8 (Bytes.unsafe_of_string buf) idx
-
 let bit_at buf i =
   let byte_num = i / 8 in
   let bit_num = i mod 8 in
-  let byte = string_get_uint8 buf byte_num in
+  let byte = String.get_uint8 buf byte_num in
   byte land (1 lsl bit_num) <> 0
 
 module type Dh = sig
   type secret
-  val secret_of_cs : ?compress:bool -> Cstruct.t ->
-    (secret * Cstruct.t, error) result
+  val secret_of_octets : ?compress:bool -> string ->
+    (secret * string, error) result
+  val secret_to_octets : secret -> string
   val gen_key : ?compress:bool -> ?g:Mirage_crypto_rng.g -> unit ->
-    secret * Cstruct.t
-  val key_exchange : secret -> Cstruct.t -> (Cstruct.t, error) result
+    secret * string
+  val key_exchange : secret -> string -> (string, error) result
 end
 
 module type Dsa = sig
   type priv
   type pub
   val byte_length : int
-  val priv_of_cstruct : Cstruct.t -> (priv, error) result
-  val priv_to_cstruct : priv -> Cstruct.t
-  val pub_of_cstruct : Cstruct.t -> (pub, error) result
-  val pub_to_cstruct : ?compress:bool -> pub -> Cstruct.t
+  val bit_length : int
+  val priv_of_octets : string -> (priv, error) result
+  val priv_to_octets : priv -> string
+  val pub_of_octets : string -> (pub, error) result
+  val pub_to_octets : ?compress:bool -> pub -> string
   val pub_of_priv : priv -> pub
   val generate : ?g:Mirage_crypto_rng.g -> unit -> priv * pub
-  val sign : key:priv -> ?k:Cstruct.t -> Cstruct.t -> Cstruct.t * Cstruct.t
-  val verify : key:pub -> Cstruct.t * Cstruct.t -> Cstruct.t -> bool
-  module K_gen (H : Mirage_crypto.Hash.S) : sig
-    val generate : key:priv -> Cstruct.t -> Cstruct.t
+  val sign : key:priv -> ?k:string -> string -> string * string
+  val verify : key:pub -> string * string -> string -> bool
+  module K_gen (H : Digestif.S) : sig
+    val generate : key:priv -> string -> string
   end
   module Precompute : sig
     val generator_tables : unit -> string array array array
@@ -85,6 +83,7 @@ module type Parameters = sig
   val n : field_element
   val pident: string
   val byte_length : int
+  val bit_length : int
   val fe_length : int
   val first_byte_bits : int option
 end
@@ -135,7 +134,7 @@ end
 module Make_field_element (P : Parameters) (F : Foreign) : Field_element = struct
   let b_uts b = Bytes.unsafe_to_string b
 
-  let create () = Bytes.make P.fe_length '\000'
+  let create () = Bytes.create P.fe_length
 
   let mul a b =
     let tmp = create () in
@@ -158,7 +157,8 @@ module Make_field_element (P : Parameters) (F : Foreign) : Field_element = struc
     b_uts tmp
 
   let zero =
-    b_uts (create ())
+    let b = Bytes.make P.fe_length '\000' in
+    b_uts b
 
   let one =
     let fe = create () in
@@ -190,7 +190,7 @@ module Make_field_element (P : Parameters) (F : Foreign) : Field_element = struc
     b_uts tmp
 
   let create_octets () =
-    Bytes.make P.byte_length '\000'
+    Bytes.create P.byte_length
 
   let to_octets fe =
     let tmp = create_octets () in
@@ -270,7 +270,7 @@ module Make_point (P : Parameters) (F : Foreign) : Point = struct
     | exception Invalid_argument _ -> None
     | false -> Some (Fe.from_be_octets buf)
 
-  (** Convert cstruct coordinates to a finite point ensuring:
+  (** Convert coordinates to a finite point ensuring:
       - x < p
       - y < p
       - y^2 = ax^3 + ax + b
@@ -307,19 +307,19 @@ module Make_point (P : Parameters) (F : Foreign) : Point = struct
       | None -> String.make 1 '\000'
       | Some (x, y) ->
         let len_x = String.length x and len_y = String.length y in
-        let res = Bytes.make (1 + len_x + len_y) '\000' in
+        let res = Bytes.create (1 + len_x + len_y) in
         Bytes.set res 0 '\004' ;
         let rev_x = rev_string x and rev_y = rev_string y in
-        Bytes.blit_string rev_x 0 res 1 len_x ;
-        Bytes.blit_string rev_y 0 res (1 + len_x) len_y ;
+        Bytes.unsafe_blit_string rev_x 0 res 1 len_x ;
+        Bytes.unsafe_blit_string rev_y 0 res (1 + len_x) len_y ;
         Bytes.unsafe_to_string res
     in
     if compress then
-      let out = Bytes.make (P.byte_length + 1) '\000' in
+      let out = Bytes.create (P.byte_length + 1) in
       let ident =
-        2 + (string_get_uint8 buf ((P.byte_length * 2) - 1)) land 1
+        2 + (String.get_uint8 buf ((P.byte_length * 2) - 1)) land 1
       in
-      Bytes.blit_string buf 1 out 1 P.byte_length;
+      Bytes.unsafe_blit_string buf 1 out 1 P.byte_length;
       Bytes.set_uint8 out 0 ident;
       Bytes.unsafe_to_string out
     else
@@ -386,15 +386,15 @@ module Make_point (P : Parameters) (F : Foreign) : Point = struct
       let y' = Fe.from_montgomery y' in
       let y_struct2 = Fe.to_octets y' in (* number must not be in montgomery domain*)
       let y_struct2 = rev_string y_struct2 in
-      let ident = string_get_uint8 pk 0 in
+      let ident = String.get_uint8 pk 0 in
       let signY =
-        2 + (string_get_uint8 y_struct (P.byte_length - 2)) land 1
+        2 + (String.get_uint8 y_struct (P.byte_length - 2)) land 1
       in
       let res = if Int.equal signY ident then y_struct else y_struct2 in
-      let out = Bytes.make ((P.byte_length * 2) + 1) '\000' in
+      let out = Bytes.create ((P.byte_length * 2) + 1) in
       Bytes.set out 0 '\004';
-      Bytes.blit_string pk 1 out 1 P.byte_length;
-      Bytes.blit_string res 0 out (P.byte_length + 1) P.byte_length;
+      Bytes.unsafe_blit_string pk 1 out 1 P.byte_length;
+      Bytes.unsafe_blit_string res 0 out (P.byte_length + 1) P.byte_length;
       Bytes.unsafe_to_string out
 
   let of_octets buf =
@@ -407,7 +407,7 @@ module Make_point (P : Parameters) (F : Foreign) : Point = struct
         let y = String.sub buf (1 + len) len in
         validate_finite_point ~x ~y
       in
-      match string_get_uint8 buf 0 with
+      match String.get_uint8 buf 0 with
       | 0x00 when String.length buf = 1 ->
         Ok (at_infinity ())
       | 0x02 | 0x03 when String.length P.pident > 0 ->
@@ -506,33 +506,23 @@ module Make_dh (Param : Parameters) (P : Point) (S : Scalar) : Dh = struct
     | Ok p -> Ok (p, share ?compress p)
     | Error _ as e -> e
 
-  let secret_of_cs ?compress s =
-    Result.map (fun (p, share) -> p, Cstruct.of_string share)
-      (secret_of_octets ?compress (Cstruct.to_string s))
+  let secret_to_octets s =
+    S.to_octets s
 
   let rec generate_private_key ?g () =
     let candidate = Mirage_crypto_rng.generate ?g Param.byte_length in
-    match S.of_octets (Cstruct.to_string candidate) with
+    match S.of_octets candidate with
     | Ok secret -> secret
     | Error _ -> generate_private_key ?g ()
 
-  let gen_key_octets ?compress ?g () =
-    let private_key = generate_private_key ?g () in
-    (private_key, share ?compress private_key)
-
   let gen_key ?compress ?g () =
-    let private_key, share = gen_key_octets ?compress ?g () in
-    private_key, Cstruct.of_string share
+    let private_key = generate_private_key ?g () in
+    private_key, share ?compress private_key
 
-  let key_exchange_octets secret received =
+  let key_exchange secret received =
     match point_of_octets received with
     | Error _ as err -> err
     | Ok shared -> Ok (P.x_of_finite_point (S.scalar_mult secret shared))
-
-  let key_exchange secret received =
-    match key_exchange_octets secret (Cstruct.to_string received) with
-    | Error _ as err -> err
-    | Ok shared -> Ok (Cstruct.of_string shared)
 end
 
 module type Foreign_n = sig
@@ -560,9 +550,9 @@ end
 module Make_Fn (P : Parameters) (F : Foreign_n) : Fn = struct
   let b_uts = Bytes.unsafe_to_string
 
-  let create () = Bytes.make P.fe_length '\000'
+  let create () = Bytes.create P.fe_length
 
-  let create_octets () = Bytes.make P.byte_length '\000'
+  let create_octets () = Bytes.create P.byte_length
 
   let from_be_octets v =
     let v' = create () in
@@ -607,26 +597,24 @@ module Make_Fn (P : Parameters) (F : Foreign_n) : Fn = struct
     b_uts tmp
 end
 
-module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Mirage_crypto.Hash.S) = struct
+module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Digestif.S) = struct
   type priv = scalar
 
   let byte_length = Param.byte_length
 
+  let bit_length = Param.bit_length
+
   let priv_of_octets= S.of_octets
 
   let priv_to_octets = S.to_octets
 
-  let priv_of_cstruct cs = priv_of_octets (Cstruct.to_string cs)
-
-  let priv_to_cstruct p = Cstruct.of_string (priv_to_octets p)
-
   let padded msg =
     let l = String.length msg in
     let bl = Param.byte_length in
     let first_byte_ok () =
       match Param.first_byte_bits with
       | None -> true
-      | Some m -> (string_get_uint8 msg 0) land (0xFF land (lnot m)) = 0
+      | Some m -> (String.get_uint8 msg 0) land (0xFF land (lnot m)) = 0
     in
     if l > bl || (l = bl && not (first_byte_ok ())) then
       raise Message_too_long
@@ -634,54 +622,51 @@ module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Mira
       msg
     else
       ( let res = Bytes.make bl '\000' in
-        Bytes.blit_string msg 0 res (bl - l) (String.length msg) ;
+        Bytes.unsafe_blit_string msg 0 res (bl - l) l ;
         Bytes.unsafe_to_string res )
 
-  let padded_cs msg =
-    let l = Cstruct.length msg in
-    let bl = Param.byte_length in
-    let first_byte_ok () =
-      match Param.first_byte_bits with
-      | None -> true
-      | Some m -> (Cstruct.get_uint8 msg 0) land (0xFF land (lnot m)) = 0
-    in
-    if l > bl || (l = bl && not (first_byte_ok ())) then
-      raise Message_too_long
-    else if l = bl then
-      msg
-    else
-      Cstruct.append (Cstruct.create (bl - l)) msg
-
   (* RFC 6979: compute a deterministic k *)
-  module K_gen (H : Mirage_crypto.Hash.S) = struct
+  module K_gen (H : Digestif.S) = struct
     let drbg : 'a Mirage_crypto_rng.generator =
       let module M = Mirage_crypto_rng.Hmac_drbg (H) in (module M)
 
-    let g ~key cs =
+    let g ~key msg =
       let g = Mirage_crypto_rng.create ~strict:true drbg in
-      Mirage_crypto_rng.reseed ~g
-        (Cstruct.append (Cstruct.of_string (S.to_octets key)) cs);
+      Mirage_crypto_rng.reseed ~g (S.to_octets key ^ msg);
       g
 
-    let g_octets ~key msg =
-      let g = Mirage_crypto_rng.create ~strict:true drbg in
-      Mirage_crypto_rng.reseed ~g
-        (Cstruct.of_string (String.concat "" [ S.to_octets key ; msg ]));
-      g
+    (* Defined in RFC 6979 sec 2.3.2 with
+       - blen = 8 * Param.byte_length
+       - qlen = Param.bit_length *)
+    let bits2int r =
+      (* keep qlen *leftmost* bits *)
+      let shift = (8 * Param.byte_length) - Param.bit_length in
+      if shift = 0 then
+        Bytes.unsafe_to_string r
+      else
+        (* Assuming shift is < 8 *)
+        let r' = Bytes.create Param.byte_length in
+        let p = ref 0x00 in
+        for i = 0 to Param.byte_length - 1 do
+          let x = Bytes.get_uint8 r i in
+          let v = (x lsr shift) lor (!p lsl (8 - shift)) in
+          p := x;
+          Bytes.set_uint8 r' i v
+        done;
+        Bytes.unsafe_to_string r'
 
     (* take qbit length, and ensure it is suitable for ECDSA (> 0 & < n) *)
     let gen g =
       let rec go () =
-        let r = Mirage_crypto_rng.generate ~g Param.byte_length in
-        let r = Cstruct.to_string r in
+        let b = Bytes.create Param.byte_length in
+        Mirage_crypto_rng.generate_into ~g b Param.byte_length;
+        (* truncate to the desired number of bits *)
+        let r = bits2int b in
         if S.is_in_range r then r else go ()
       in
       go ()
 
-    (* let generate_octets ~key buf = gen (g ~key (Cstruct.of_string (padded buf))) *)
-
-    let generate ~key buf =
-      Cstruct.of_string (gen (g ~key (padded_cs buf)))
+    let generate ~key buf = gen (g ~key (padded buf))
   end
 
   module K_gen_default = K_gen(H)
@@ -692,16 +677,11 @@ module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Mira
 
   let pub_to_octets ?(compress = false) pk = P.to_octets ~compress pk
 
-  let pub_of_cstruct cs = pub_of_octets (Cstruct.to_string cs)
-
-  let pub_to_cstruct ?compress p =
-    Cstruct.of_string (pub_to_octets ?compress p)
-
   let generate ?g () =
     (* FIPS 186-4 B 4.2 *)
     let d =
       let rec one () =
-        match S.of_octets (Cstruct.to_string (Mirage_crypto_rng.generate ?g Param.byte_length)) with
+        match S.of_octets (Mirage_crypto_rng.generate ?g Param.byte_length) with
         | Ok x -> x
         | Error _ -> one ()
       in
@@ -719,10 +699,10 @@ module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Mira
       let x = F.from_montgomery x in
       Some (F.to_be_octets x)
 
-  let sign_octets ~key ?k msg =
+  let sign ~key ?k msg =
     let msg = padded msg in
     let e = F.from_be_octets msg in
-    let g = K_gen_default.g_octets ~key msg in
+    let g = K_gen_default.g ~key msg in
     let rec do_sign g =
       let again () =
         match k with
@@ -754,13 +734,9 @@ module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Mira
     in
     do_sign g
 
-  let sign ~key ?k msg =
-    let r, s = sign_octets ~key ?k:(Option.map Cstruct.to_string k) (Cstruct.to_string msg) in
-    Cstruct.of_string r, Cstruct.of_string s
-
   let pub_of_priv priv = S.scalar_mult_base priv
 
-  let verify_octets ~key (r, s) msg =
+  let verify ~key (r, s) msg =
     try
       let r = padded r and s = padded s in
       if not (S.is_in_range r && S.is_in_range s) then
@@ -793,64 +769,11 @@ module Make_dsa (Param : Parameters) (F : Fn) (P : Point) (S : Scalar) (H : Mira
     with
     | Message_too_long -> false
 
-  let verify ~key (r, s) digest =
-    verify_octets ~key (Cstruct.to_string r, Cstruct.to_string s) (Cstruct.to_string digest)
-
   module Precompute = struct
     let generator_tables = S.generator_tables
   end
 end
 
-module P224 : Dh_dsa = struct
-  module Params = struct
-    let a = "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFE\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFE"
-    let b = "\xB4\x05\x0A\x85\x0C\x04\xB3\xAB\xF5\x41\x32\x56\x50\x44\xB0\xB7\xD7\xBF\xD8\xBA\x27\x0B\x39\x43\x23\x55\xFF\xB4"
-    let g_x = "\xB7\x0E\x0C\xBD\x6B\xB4\xBF\x7F\x32\x13\x90\xB9\x4A\x03\xC1\xD3\x56\xC2\x11\x22\x34\x32\x80\xD6\x11\x5C\x1D\x21"
-    let g_y = "\xBD\x37\x63\x88\xB5\xF7\x23\xFB\x4C\x22\xDF\xE6\xCD\x43\x75\xA0\x5A\x07\x47\x64\x44\xD5\x81\x99\x85\x00\x7E\x34"
-    let p = "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"
-    let n = "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\x16\xA2\xE0\xB8\xF0\x3E\x13\xDD\x29\x45\x5C\x5C\x2A\x3D"
-    let pident = ""
-    let byte_length = 28
-    let fe_length = if Sys.word_size == 64 then 32 else 28 (* TODO: is this congruent with C code? *)
-    let first_byte_bits = None
-  end
-
-  module Foreign = struct
-    external mul : out_field_element -> field_element -> field_element -> unit = "mc_p224_mul" [@@noalloc]
-    external sub : out_field_element -> field_element -> field_element -> unit = "mc_p224_sub" [@@noalloc]
-    external add : out_field_element -> field_element -> field_element -> unit = "mc_p224_add" [@@noalloc]
-    external to_montgomery : out_field_element -> field_element -> unit = "mc_p224_to_montgomery" [@@noalloc]
-    external from_octets : out_field_element -> string -> unit = "mc_p224_from_bytes" [@@noalloc]
-    external set_one : out_field_element -> unit = "mc_p224_set_one" [@@noalloc]
-    external nz : field_element -> bool = "mc_p224_nz" [@@noalloc]
-    external sqr : out_field_element -> field_element -> unit = "mc_p224_sqr" [@@noalloc]
-    external from_montgomery : out_field_element -> field_element -> unit = "mc_p224_from_montgomery" [@@noalloc]
-    external to_octets : bytes -> field_element -> unit = "mc_p224_to_bytes" [@@noalloc]
-    external inv : out_field_element -> field_element -> unit = "mc_p224_inv" [@@noalloc]
-    external select_c : out_field_element -> bool -> field_element -> field_element -> unit = "mc_p224_select" [@@noalloc]
-    external double_c : out_point -> point -> unit = "mc_p224_point_double" [@@noalloc]
-    external add_c : out_point -> point -> point -> unit = "mc_p224_point_add" [@@noalloc]
-    external scalar_mult_base_c : out_point -> string -> unit = "mc_p224_scalar_mult_base" [@@noalloc]
-  end
-
-  module Foreign_n = struct
-    external mul : out_field_element -> field_element -> field_element -> unit = "mc_np224_mul" [@@noalloc]
-    external add : out_field_element -> field_element -> field_element -> unit = "mc_np224_add" [@@noalloc]
-    external inv : out_field_element -> field_element -> unit = "mc_np224_inv" [@@noalloc]
-    external one : out_field_element -> unit = "mc_np224_one" [@@noalloc]
-    external from_bytes : out_field_element -> string -> unit = "mc_np224_from_bytes" [@@noalloc]
-    external to_bytes : bytes -> field_element -> unit = "mc_np224_to_bytes" [@@noalloc]
-    external from_montgomery : out_field_element -> field_element -> unit = "mc_np224_from_montgomery" [@@noalloc]
-    external to_montgomery : out_field_element -> field_element -> unit = "mc_np224_to_montgomery" [@@noalloc]
-  end
-
-  module P = Make_point(Params)(Foreign)
-  module S = Make_scalar(Params)(P)
-  module Dh = Make_dh(Params)(P)(S)
-  module Fn = Make_Fn(Params)(Foreign_n)
-  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Mirage_crypto.Hash.SHA256)
-end
-
 module P256 : Dh_dsa  = struct
   module Params = struct
     let a = "\xFF\xFF\xFF\xFF\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFC"
@@ -861,6 +784,7 @@ module P256 : Dh_dsa  = struct
     let n = "\xFF\xFF\xFF\xFF\x00\x00\x00\x00\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xBC\xE6\xFA\xAD\xA7\x17\x9E\x84\xF3\xB9\xCA\xC2\xFC\x63\x25\x51"
     let pident = "\x3F\xFF\xFF\xFF\xC0\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" |> rev_string (* (Params.p + 1) / 4*)
     let byte_length = 32
+    let bit_length = 256
     let fe_length = 32
     let first_byte_bits = None
   end
@@ -898,7 +822,7 @@ module P256 : Dh_dsa  = struct
   module S = Make_scalar(Params)(P)
   module Dh = Make_dh(Params)(P)(S)
   module Fn = Make_Fn(Params)(Foreign_n)
-  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Mirage_crypto.Hash.SHA256)
+  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Digestif.SHA256)
 end
 
 module P384 : Dh_dsa = struct
@@ -912,6 +836,7 @@ module P384 : Dh_dsa = struct
     let n = "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xC7\x63\x4D\x81\xF4\x37\x2D\xDF\x58\x1A\x0D\xB2\x48\xB0\xA7\x7A\xEC\xEC\x19\x6A\xCC\xC5\x29\x73"
     let pident = "\x3F\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xBF\xFF\xFF\xFF\xC0\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00" |> rev_string (* (Params.p + 1) / 4*)
     let byte_length = 48
+    let bit_length = 384
     let fe_length = 48
     let first_byte_bits = None
   end
@@ -949,7 +874,7 @@ module P384 : Dh_dsa = struct
   module S = Make_scalar(Params)(P)
   module Dh = Make_dh(Params)(P)(S)
   module Fn = Make_Fn(Params)(Foreign_n)
-  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Mirage_crypto.Hash.SHA384)
+  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Digestif.SHA384)
 end
 
 module P521 : Dh_dsa = struct
@@ -964,6 +889,7 @@ module P521 : Dh_dsa = struct
     let n = "\x01\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFA\x51\x86\x87\x83\xBF\x2F\x96\x6B\x7F\xCC\x01\x48\xF7\x09\xA5\xD0\x3B\xB5\xC9\xB8\x89\x9C\x47\xAE\xBB\x6F\xB7\x1E\x91\x38\x64\x09"
     let pident = "\x01\x7f\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff" |> rev_string
     let byte_length = 66
+    let bit_length = 521
     let fe_length = if Sys.word_size == 64 then 72 else 68  (* TODO: is this congruent with C code? *)
     let first_byte_bits = Some 0x01
   end
@@ -1001,7 +927,7 @@ module P521 : Dh_dsa = struct
   module S = Make_scalar(Params)(P)
   module Dh = Make_dh(Params)(P)(S)
   module Fn = Make_Fn(Params)(Foreign_n)
-  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Mirage_crypto.Hash.SHA512)
+  module Dsa = Make_dsa(Params)(Fn)(P)(S)(Digestif.SHA512)
 end
 
 module X25519 = struct
@@ -1011,7 +937,7 @@ module X25519 = struct
   let key_len = 32
 
   let scalar_mult in_ base =
-    let out = Bytes.make key_len '\000' in
+    let out = Bytes.create key_len in
     x25519_scalar_mult_generic out in_ base;
     Bytes.unsafe_to_string out
 
@@ -1021,38 +947,28 @@ module X25519 = struct
 
   let public priv = scalar_mult priv basepoint
 
-  let gen_key_octets ?compress:_ ?g () =
-    let secret = Cstruct.to_string (Mirage_crypto_rng.generate ?g key_len) in
+  let gen_key ?compress:_ ?g () =
+    let secret = Mirage_crypto_rng.generate ?g key_len in
     secret, public secret
 
-  let gen_key ?compress ?g () =
-    let secret, public = gen_key_octets ~compress ?g () in
-    secret, Cstruct.of_string public
-
   let secret_of_octets ?compress:_ s =
     if String.length s = key_len then
       Ok (s, public s)
     else
       Error `Invalid_length
 
-  let secret_of_cs ?compress cs =
-    Result.map (fun (secret, public) -> secret, Cstruct.of_string public)
-      (secret_of_octets ~compress (Cstruct.to_string cs))
+  let secret_to_octets s = s
 
   let is_zero =
     let zero = String.make key_len '\000' in
     fun buf -> String.equal zero buf
 
-  let key_exchange_octets secret public =
+  let key_exchange secret public =
     if String.length public = key_len then
       let res = scalar_mult secret public in
       if is_zero res then Error `Low_order else Ok res
     else
       Error `Invalid_length
-
-  let key_exchange secret public =
-    Result.map Cstruct.of_string
-      (key_exchange_octets secret (Cstruct.to_string public))
 end
 
 module Ed25519 = struct
@@ -1065,17 +981,17 @@ module Ed25519 = struct
   let key_len = 32
 
   let scalar_mult_base_to_bytes p =
-    let tmp = Bytes.make key_len '\000' in
+    let tmp = Bytes.create key_len in
     scalar_mult_base_to_bytes tmp p;
     Bytes.unsafe_to_string tmp
 
   let muladd a b c =
-    let tmp = Bytes.make key_len '\000' in
+    let tmp = Bytes.create key_len in
     muladd tmp a b c;
     Bytes.unsafe_to_string tmp
 
   let double_scalar_mult a b c =
-    let tmp = Bytes.make key_len '\000' in
+    let tmp = Bytes.create key_len in
     let s = double_scalar_mult tmp a b c in
     s, Bytes.unsafe_to_string tmp
 
@@ -1083,14 +999,23 @@ module Ed25519 = struct
 
   type priv = string
 
+  let sha512 datas =
+    let open Digestif.SHA512 in
+    let buf = Bytes.create digest_size in
+    let ctx = List.fold_left (feed_string ?off:None ?len:None) empty datas in
+    get_into_bytes ctx buf;
+    buf
+
   (* RFC 8032 *)
   let public secret =
     (* section 5.1.5 *)
     (* step 1 *)
-    let h = Mirage_crypto.Hash.SHA512.digest (Cstruct.of_string secret) in
+    let h = sha512 [ secret ] in
     (* step 2 *)
-    let s, rest = Cstruct.split h key_len in
-    let s, rest = Cstruct.to_bytes s, Cstruct.to_string rest in
+    let s, rest =
+      Bytes.sub h 0 key_len,
+      Bytes.unsafe_to_string (Bytes.sub h key_len (Bytes.length h - key_len))
+    in
     Bytes.set_uint8 s 0 ((Bytes.get_uint8 s 0) land 248);
     Bytes.set_uint8 s 31 (((Bytes.get_uint8 s 31) land 127) lor 64);
     let s = Bytes.unsafe_to_string s in
@@ -1103,11 +1028,7 @@ module Ed25519 = struct
   let priv_of_octets buf =
     if String.length buf = key_len then Ok buf else Error `Invalid_length
 
-  let priv_of_cstruct p = priv_of_octets (Cstruct.to_string p)
-
-  let priv_to_octets priv = priv
-
-  let priv_to_cstruct p = Cstruct.of_string (priv_to_octets p)
+  let priv_to_octets (priv : priv) = priv
 
   let pub_of_octets buf =
     if String.length buf = key_len then
@@ -1118,38 +1039,29 @@ module Ed25519 = struct
     else
       Error `Invalid_length
 
-  let pub_of_cstruct p = pub_of_octets (Cstruct.to_string p)
-
   let pub_to_octets pub = pub
 
-  let pub_to_cstruct p = Cstruct.of_string (pub_to_octets p)
-
   let generate ?g () =
     let secret = Mirage_crypto_rng.generate ?g key_len in
-    let secret = Cstruct.to_string secret in
     secret, pub_of_priv secret
 
-  let sign_octets ~key msg =
+  let sign ~key msg =
     (* section 5.1.6 *)
     let pub, (s, prefix) = public key in
-    let r = Mirage_crypto.Hash.SHA512.digest (Cstruct.of_string (String.concat "" [ prefix; msg ])) in
-    let r = Cstruct.to_bytes r in
+    let r = sha512 [ prefix; msg ] in
     reduce_l r;
     let r = Bytes.unsafe_to_string r in
     let r_big = scalar_mult_base_to_bytes r in
-    let k = Mirage_crypto.Hash.SHA512.digest (Cstruct.of_string (String.concat "" [ r_big; pub; msg])) in
-    let k = Cstruct.to_bytes k in
+    let k = sha512 [ r_big; pub; msg] in
     reduce_l k;
     let k = Bytes.unsafe_to_string k in
     let s_out = muladd k s r in
-    let res = Bytes.make (key_len + key_len) '\000' in
-    Bytes.blit_string r_big 0 res 0 key_len ;
-    Bytes.blit_string s_out 0 res key_len key_len ;
+    let res = Bytes.create (key_len + key_len) in
+    Bytes.unsafe_blit_string r_big 0 res 0 key_len ;
+    Bytes.unsafe_blit_string s_out 0 res key_len key_len ;
     Bytes.unsafe_to_string res
 
-  let sign ~key msg = Cstruct.of_string (sign_octets ~key (Cstruct.to_string msg))
-
-  let verify_octets ~key signature ~msg =
+  let verify ~key signature ~msg =
     (* section 5.1.7 *)
     if String.length signature = 2 * key_len then
       let r, s =
@@ -1159,18 +1071,14 @@ module Ed25519 = struct
       let s_smaller_l =
         (* check s within 0 <= s < L *)
         let s' = Bytes.make (key_len * 2) '\000' in
-        Bytes.blit_string s 0 s' 0 key_len;
+        Bytes.unsafe_blit_string s 0 s' 0 key_len;
         reduce_l s';
         let s' = Bytes.unsafe_to_string s' in
-        let s'' = String.concat "" [ s; String.make key_len '\000' ] in
+        let s'' = s ^ String.make key_len '\000' in
         String.equal s'' s'
       in
       if s_smaller_l then begin
-        let k =
-          let data_to_hash = String.concat "" [ r ; key ; msg ] in
-          Mirage_crypto.Hash.SHA512.digest (Cstruct.of_string data_to_hash)
-        in
-        let k = Cstruct.to_bytes k in
+        let k = sha512 [ r ; key ; msg ] in
         reduce_l k;
         let k = Bytes.unsafe_to_string k in
         let success, r' = double_scalar_mult k key s in
@@ -1179,7 +1087,4 @@ module Ed25519 = struct
         false
     else
       false
-
-  let verify ~key signature ~msg =
-    verify_octets ~key (Cstruct.to_string signature) ~msg:(Cstruct.to_string msg)
 end
