@@ -16,16 +16,22 @@
 
 (** {b TL;DR} Don't forget to seed; don't maintain your own [g].
 
-    The RNGs here are merely the deterministic part of a full random number
-    generation suite. For proper operation, they need to be seeded with a
-    high-quality entropy source.
+    For common operations on Unix (independent of your asynchronous task
+    library, you can use /dev/urandom or getentropy(3) (actually getrandom(3) on
+    Linux, getentropy() on macOS and BSD systems, BCryptGenRandom on Windows).
+
+    Please ensure to call [Mirage_crypto_rng_unix.use_default], or
+    [Mirage_crypto_rng_unix.use_dev_urandom] (if you only want to use
+    /dev/urandom), or [Mirage_crypto_rng_unix.use_getentropy] (if you only want
+    to use getrandom/getentropy/BCryptGenRandom).
+
+    For fine-grained control (doing entropy harvesting, etc.), please continue
+    reading the documentation below. {b Please be aware that the feeding of
+    Fortuna and producing random numbers is not thread-safe} (it is on Miou_unix
+    via Pfortuna).
 
     Suitable entropy feeding of generators are provided by other libraries
-    {{!Mirage_crypto_rng_lwt}mirage-crypto-rng-lwt} (for Lwt),
-    {{!Mirage_crypto_rng_async}mirage-crypto-rng-async} (for Async),
     {{!Mirage_crypto_rng_mirage}mirage-crypto-rng-mirage} (for MirageOS),
-    {{!Mirage_crypto_rng_unix}mirage-crypto-rng.unix},
-    {{!Mirage_crypto_rng_eio}mirage-crypto-rng-eio} (for Eio),
     and {{!Mirage_crypto_rng_miou_unix}mirage-crypto-miou-unix} (for Miou_unix).
 
     The intention is that "initialize" in the respective sub-library is called
@@ -36,9 +42,9 @@
     generator should be used in most setting, and that should be fed a constant
     stream of entropy.
 
-    [mirage-crypto-rng-eio] package differs slightly from other rng packages.
-    Instead of the [initialize] function a [run] function is provided with
-    similar behaviour, i.e. RNG setup, entropy collection and periodic reseeding.
+    The RNGs here are merely the deterministic part of a full random number
+    generation suite. For proper operation, they need to be seeded with a
+    high-quality entropy source.
 
     Although this module exposes a more fine-grained interface, e.g. allowing
     manual seeding of generators, this is intended either for implementing
@@ -101,11 +107,13 @@ module Entropy : sig
 
   val cpu_rng_bootstrap : (int -> string, [`Not_supported]) Result.t
   (** [cpu_rng_bootstrap id] returns 8 bytes of random data using the CPU
-      RNG (rdseed or rdrand). On 32bit platforms, only 4 bytes are filled.
-      The [id] is used as prefix.
+      RNG (rdseed). On 32bit platforms, only 4 bytes are filled.
+      The [id] is used as prefix. If only rdrand is available, the return
+      value is the concatenation of 512 calls to rdrand.
 
-      @raise Failure if no CPU RNG is available, or if it doesn't return a
-      random value. *)
+      @raise Failure if rdrand fails 512 times, or if rdseed fails and rdrand
+      is not available.
+  *)
 
   val bootstrap : int -> string
   (** [bootstrap id] is either [cpu_rng_bootstrap], if the CPU supports it, or
@@ -123,14 +131,26 @@ module Entropy : sig
 
   (** {1 Periodic pulled sources} *)
 
-  val feed_pools : g option -> source -> (unit -> string) -> unit
+  val feed_pools : g option -> source -> (unit -> (string, [ `No_random_available ]) result) -> unit
   (** [feed_pools g source f] feeds all pools of [g] using [source] by executing
       [f] for each pool. *)
 
   val cpu_rng : (g option -> unit -> unit, [`Not_supported]) Result.t
   (** [cpu_rng g] uses the CPU RNG (rdrand or rdseed) to feed all pools
       of [g]. It uses {!feed_pools} internally. If neither rdrand nor rdseed
-      are available, [fun () -> ()] is returned. *)
+      are available, [`Not_supported] is returned. *)
+
+  val rdrand_calls : unit -> int
+  (** [rdrand_calls ()] returns the number of rdrand calls. *)
+
+  val rdrand_failures : unit -> int
+  (** [rdrand_failures ()] returns the number of rdrand failures. *)
+
+  val rdseed_calls : unit -> int
+  (** [rdseed_calls ()] returns the number of rdseed calls. *)
+
+  val rdseed_failures : unit -> int
+  (** [rdseed_failures ()] returns the number of rdseed failures. *)
 
   (**/**)
   val id : source -> int
@@ -156,10 +176,13 @@ module type Generator = sig
   (** Create a new, unseeded {{!g}g}. *)
 
   val generate_into : g:g -> bytes -> off:int -> int -> unit
+  [@@alert unsafe "Does not do bounds checks. Use Mirage_crypto_rng.generate_into instead."]
   (** [generate_into ~g buf ~off n] produces [n] uniformly distributed random
       bytes into [buf] at offset [off], updating the state of [g].
 
-      @raise Invalid_argument if buffer is too small (it must be: Bytes.length buf - off >= n)
+      Assumes that [buf] is at least [off + n] bytes long. Also assumes that
+      [off] and [n] are positive integers. Caution: do not use in your
+      application, use [Mirage_crypto_rng.generate_into] instead.
   *)
 
   val reseed : g:g -> string -> unit
@@ -233,7 +256,11 @@ val generate_into : ?g:g -> bytes -> ?off:int -> int -> unit
 (** [generate_into ~g buf ~off len] invokes
     {{!Generator.generate_into}generate_into} on [g] or
     {{!generator}default generator}. The random data is put into [buf] starting
-    at [off] (defaults to 0) with [len] bytes. *)
+    at [off] (defaults to 0) with [len] bytes.
+
+    @raise Invalid_argument if buffer is too small (it must be: [Bytes.length
+      buf - off >= n]) or [off] or [n] are negative.
+*)
 
 val generate : ?g:g -> int -> string
 (** Invoke {!generate_into} on [g] or {{!generator}default generator} and a
